%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FLDL to Tokio Converter
%  (Very Simple Version....)
%          Wed Jul 26 11:55:44 JST 1989
% Prolog Part
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-[ts].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Top level
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
main(InFile,OutFile) :-
	see(InFile),tell(OutFile),
	read(X),
	translate(X,Y,Top,N),
	read(Z),
	main_loop(Z,Y,N,Top),
        seen,told.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% main_loop(FLDL, PreviousTokio, 
%               CurrentState,TopState)
%   one cluase delay write,
%   for fixing next status default
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
main_loop(X,P,Current,T):-   % fix prev next status
	translate(X,Y,Current,Next),!,  
	write_clause(P),     % write previous
	read(Z),
	main_loop(Z,Y,Next,T).
main_loop(_,P,Top,Top):-!,
	write_clause(P).

	   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% One Line Translate FLDL to Tokio
% tanslate(FLDL, Tokio, CurrentState,NextState)
% fail on end_of_file
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
translate(end_of_file, _,_,_) :- !,fail.
translate((H:-B),Out,Current,N) :- 
	translate1(H,B,Out,N),
	functor(H,Current,_),!.
translate(_,O,C,N) :- % skip atomic
	read(X),translate(X,O,C,N).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% One Line Translate FLDL to Tokio
% tanslate1(Head, Body, Tokio, NextState(=DefaultNext))
% Fail on No output
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
translate1(name,_,_,_):-!,fail.
translate1(date,_,_,_):-!,fail.
translate1(designer,_,_,_):-!,fail.
translate1(revision,_,_,_):-!,fail.
% translate(andin,InputVarlist,_):-
%      :- write((:-global(InputVarlist))),!,fail.
translate1(andin,InputVarlist,_,_) :-
	write_input(InputVarlist),!,fail.
translate1(orout,OutVarlist,_,_) :-
	write_output(OutVarlist),!,fail.
% Next: current status default goto statement folder ( true or @next)
% Nexts: Next list in branch status ( if or wait )
% Default: Next Status Name ( undefined now and fix later)
translate1(Atom,Body,(Atom :- Body1,Next1),Default) :-
	atom(Atom),!,
	body(Body,Body1,default(Default),Next,Atom),
	fix_next_top(Next,default(Default),Next1).
translate1(Functor,B,B1,Default) :-      %%% ignore state value 
	functor(Functor,Head,N),N=\=0,!,
	translate1(Head,B,B1,Default).
translate1(Atom,Body,Output,Default) :-!,
	trace,
	translate1(Atom,Body,Output,Default).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FLDL -> Tokio Body Conversoin
%	body(Body,Body1,OldNextStatus,
%            NewNextStatus,CurrentStatus)
%
% For each branch, there may be one possible next transition.
% If next transition exists, it propergates all default branch.
% If no transition exists, generate default transition.
% Never Fail
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
body(Atomic,Atomic,N,N,_) :- atomic(Atomic),!.
body({B},B1,Next,Next1,Self) :- !,body(B,B1,Next,Next1,Self).
body((A,B),(A1,B1),Next,Next1,Self) :- 
	body(A,A1,Next,Next2,Self),   % possible Next conflict
	body(B,B1,Next2,Next1,Self).
body((if Cond then Yes else No),
	{if Cond1 then (Yes1,NextY1) else (No1,NextN1)},
	Next,Next1,Self) :-!,
	cond(Cond,Cond1),
	body(Yes,Yes1,Next,NextY,Self),
	body(No,No1,Next,NextN,Self),
	fix_next([(NextY,NextY1),(NextN,NextN1)],Next,Next1).
body((if Cond then Yes),
	{if Cond1 then (Yes1,NextY1) else NextN1},
	      Next,Next1,Self) :-!,
	cond(Cond,Cond1),
	body(Yes,Yes1,Next,NextY,Self),
	fix_next([(NextY,NextY1),(Next,NextN1)],Next,Next1).
% It is better to check goto conflict here....
% body(goto @ next,@ Next,_,@ Next,_) :- !.
body(goto @ next, true,Default,Default,_) :- !.
body(goto @ self,@ Self,_,@ Self,Self) :- !.
body(goto Next ,@ Next,_,@ Next, _) :- !.
body((wait Cond then Exit and Loop),Out,Next,Next1,Self) :- !,
	body((if Cond then Exit else (Loop, goto @ self)),
	Out,Next,Next1,Self).
body((wait Cond then Exit),Out,Next,Next1,Self) :- !,
	body((if Cond then Exit else goto @ self),Out,Next,Next1,Self).
body((wait Cond),Out,Next,Next1,Self) :- !,
	body((if Cond then goto @ next else goto @ self),Out,Next,Next1,Self).
body(case(Pattern,Cases),Out,Next,Next1,Self) :- !,
	functor(Pattern,_,N),
	case(Cases,Pattern,N,IfThenElse),
	body(IfThenElse,Out,Next,Next1,Self).
body(Error,error, _,_) :-!,
	write(error_on(Error)),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% case(pattern(....),[pattern(....):body, .... ,end])
%  -> if (....) then body
%     else if (....) then body
%     else @ next
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
case([otherwise:Body|_],Pattern,N,Body):-!.
case([end],Pattern,N,true):-!.
case([Pattern:Body|T],Pattern0,N,(if Cond then Body else Else)) :-!,
	pattern(0,N,Pattern,Pattern0,Cond),
	case(T,Pattern0,N,Else).

pattern(N,N,_,_,true) :-!.
pattern(N1,N,P,P0,(Cond,Cond1)) :-
	N2 is N1+1,arg(N2,P,A),
	pattern1(A,N2,P,P0,Cond),
	pattern(N2,N,P,P0,Cond1).
pattern1(V,_,_,P,P0,true) :- var(V),!.
pattern1(1,N1,P,P0,Cond) :- !,
	arg(N1,P0,Cond).
pattern1(0,N1,P,P0,not(Cond)) :- !,
	arg(N1,P0,Cond).

cond(true,true) :-!.
cond((A,B),(A1,B1)) :- !,cond(A,A1),cond(B,B1).
cond(A&B,(A1,B1)) :- !,cond(A,A1),cond(B,B1).
cond(not(Atom),not(*Atom)):- atom(Atom),!.
cond(not(Exp),not(Exp1)):- !,cond(Exp,Exp1).
cond(Atom,*Atom):- !,atom(Atom).
cond(Error,error) :- 
	write('Unknown Condition Error on '),write(Error),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set default next state
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% If all Nexts are default then
%       set all subnexts = true and set outnext = Default
% else
%       fill all other subnext's defaults as @ Default 
%       and set outnext = true
% Never Fail
fix_next(Nexts,Default,Default) :-
	fix_on_alldefault(Nexts),!.
fix_next(Nexts,Default,true) :-
	fix_on_some_goto(Nexts,Default),!.
fix_next(Nexts,Default,true) :-
	write('Error on fix_next:'),
	write(fix_next(Nexts,Default,true)),nl.

fix_on_alldefault([(default(_),true)|T]) :-!,
	fix_on_alldefault(T).
fix_on_alldefault([]).

fix_on_some_goto([],Default).
fix_on_some_goto([(default(_),@ Default)|T],default(Default)) :-!,
	fix_on_some_goto(T,default(Default)).
fix_on_some_goto([(NoDefault,true)|T],Default) :-!,
	fix_on_some_goto(T,Default).

fix_next_top(default(N),default(N),@ N):-!.
fix_next_top(_,Next,true).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Simple Output List Predicates
%    out -->  out :- write(' out')
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_output1(H) :- (H = +X;H = -X;H = X),
	write((X :- write(H))),put("."),nl.
write_output((H,T)):-!,
	write_output1(H),
	write_output(T).
write_output(H):-
	write_output1(H).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Simple Input  Predicates
%    in -->  in :- *in = 1,
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_input1(H) :- 
	write((H :- *H = 1)),put("."),nl.
write_input((H,T)):-!,
	write_input1(H),
	write_input(T).
write_input(H):-
	write_input1(H).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Interface Utilities
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_clause(true):-!.
write_clause(end_of_file):-!.
write_clause(B) :- 
	write_clause1(B),fail.
write_clause(_).

write_clause1((X:-B)) :-!,
	c_opt(B,B1,_),
	numbervars((X:-B1),0,_),
	pretty((X:-B1)).
%	write((X:-B1)),put("."),nl,!.
write_clause1(Y) :-
	numbervars(Z,0,_),
	write(Z),put("."),nl,!.

% c_opt(Before, After, Optimized_or_not) 
c_opt(X,X,0) :- var(X),!.
c_opt({X},{X1},0) :- !,c_opt(X,X1,_).
c_opt((if A then B else true),(if A1 then B1),0) :-!,
	c_opt(A,A1,_),	c_opt(B,B1,_).
c_opt((if A then B else C),(if A1 then B1 else C1),0) :-!,
	c_opt(A,A1,_),	c_opt(B,B1,_),	c_opt(C,C1,_).
c_opt((true, X), Y, 1) :- !, c_opt(X, Y, _). 
c_opt((X0, true), X, 1) :- !, c_opt(X0, X, _). 
c_opt((X0, Y0), R, C) :- !, 
    c_opt(X0, X1, CX), c_opt(Y0, Y1, CY), C0 is CX\/CY, 
  ( C0 =:= 0, !, C is C0, R = (X1, Y1);  
    c_opt((X1, Y1), R, C) ).            % X1/Y1 may be optimized to 'true' 
c_opt(X, X, 0). 

% end %

