%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pretty Printer for
%   Tokio from FLDL
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% :-[ts].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Top level
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% main(InFile) :- main(InFile,user).
% main(InFile,OutFile) :-
%	see(InFile),tell(OutFile),
%	repeat,
%		read(X),pretty(X),
%		X = end_of_file,
%        seen,told.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pretty
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pretty(end_of_file) :- !.
pretty(Atom) :- atomic(Atom),
	write(Atom),put("."),nl,!,fail.
pretty((H:-true)) :- !,
	write(H),put("."),nl,!,fail.
pretty((H:-B)) :- !,nl,
	write(H),write(' :- '),
	prbody0(B,4),put("."),nl,
	!,fail.

% no need of {}
prbody0({A},I) :- !,prbody(A,I).
prbody0(A,I) :- prbody(A,I).

prbody((A,B),I) :- !,
	prbody(A,I),put(","),prbody(B,I).
prbody((if A then B else C),I) :- !,
	nl,tab(I),write('if '),J is I + 4,
	prcond(A,J),write(' then '),
	prbody(B,J),nl,
	tab(I),write('else '),
	elsif(C,I,J).
prbody((if A then B),I) :- !,
	nl,tab(I),write('if '),J is I + 4,
	prcond(A,J),write(' then '),
	prbody(B,J).
prbody({A},I) :- !,
	write('{'),
	prbody0(A,I),
	write('}').
prbody(A,I) :-
	nl,tab(I),write(A).
elsif((if A then B else C),I,J) :-!,
	write('if '),
	prcond(A,J),write(' then '),
	prbody(B,J),nl,
	tab(I),write('else '),
	elsif(C,I,J).
elsif((if A then B),I,J) :-!,
	write('if '),
	prcond(A,J),write(' then '),
	prbody(B,J).
elsif({C},I,J) :-
	elsif(C,I,J).
elsif(C,I,J) :-
	prbody(C,J).

prcond(A,_) :- write(A).
