%
%	time tracer    backward
%
%		problem : subinterval that calls predicates parallelly
%				must determine minimum length
%
%
sameTimeStateCheck :-
	!,
	show_record(conflict),
	show_record(stateConflict),
	getSTstates,
	show_record(stateSet),
	mayHappenST.

getSTstates :-        %  hontou ha douji deha nai ga isshoni hyouka site yoi
	recorded(stateConflict,(S1,S2),Ref),
	getSTstatesMain(S1,S2),
%	erase(Ref),
	fail.
getSTstates :-
	!.

getSTstatesMain(S1,S2) :-
	recorded(stateSet,Set,Ref),
	memberOr(S1,S2,Set,NS),
	!,
	erase(Ref),
	recordz(stateSet,NS,_).
getSTstatesMain(S1,S2) :-
	!,
	recordz(stateSet,[S1,S2],_).

memberOr(S1,S2,Set,NS) :-
	member(S1,Set),
	!,
	memberOr2(S2,Set,NS).
memberOr(S1,S2,Set,NS) :-
	member(S2,Set),
	!,
	memberOr2(S1,Set,NS).

memberOr2(S,Set,Set) :-
	member(S,Set),
	!.
memberOr2(S,Set,[S|Set]) :-
	!.

mayHappenST:-
	recorded(stateSet,Set,_),
	write('This subinterval set happens at same time.'),write(Set),nl,
	mayHappenSTMain(Set,[],St),	
	statusPrint(Set,St),
	fail.
mayHappenST:-
	!.

mayHappenSTMain([],Al,success):-
	!.
mayHappenSTMain([State|Set],Al,St):-
	recorded(facility,(State,Id,_),_),
	notConflictRec((State,Id),Al),
	mayHappenSTMain(Set,[(State,Id)|Al],St),
	!.
mayHappenSTMain(_,_,fail):-
	!.

notConflictRec(FS,[]) :-
	!.
notConflictRec(FS,[A|AA]) :-
	recordAsConflict(FS,A),
	!,
	fail.
notConflictRec(FS,[A|AA]) :-
	notConflictRec(FS,AA),
	!.

recordAsConflict(S1,S2) :-
	recorded(conflict,(S1,S2),_),
	!.
recordAsConflict(S1,S2) :-
	recorded(conflict,(S2,S1),_),
	!.
	
statusPrint(Set,success) :-
	!,
	write('Can. This set can be implemented.'),write(Set),nl.
statusPrint(Set,fail) :-
	!,
	write('Not. This set can not be implemented.'),write(Set),nl,
	write('Conflict happens in some facilities.'),nl.

allTrace :-
	clear_record(sameTimeHappen),
	clear_record(notHappen),
	clear_record(dependOn),
	clear_record(already),
	conflictSet(FS1,FS2,Conf),
%	display(('New conflict set has gotten.')),ttynl,
%	display(('New conflict set has gotten.      ')),
	write(FS1),nl,write(' and '),nl,write(FS2),nl,
	write(' may conflict in '),nl,write(Conf),nl,nl,
	backTrace(FS1,FS2),
%	display(('Finished.')),ttynl,
	fail.
allTrace :-
	show_record(sameTimeHappen),
	show_record(notHappen),
	show_record(dependOn).

backTrace(FS1,FS2) :-
	FS1 = (S1,Id1),
	FS2 = (S2,Id2),
	sameStateHandle(S1,T1,C1),
	sameStateHandle(S2,T2,C2),
	notExCond(C1,C2),
%	display('  happenSameTime start.'),ttynl,
	happenSameTime(T1,T2,[],[(T1,S1)],[(T2,S2)],Pr1,Pr2),
	!,
%	display('  happenRecord start.'),ttynl,
	show_record(dependOn),
	clear_record(appear),
	happenRecord,
	printProcess(Pr1,0),nl,
	printProcess(Pr2,0),nl,
	recordz(conflict,(FS1,FS2),_),
	recordz(stateConflict,(S1,S2),_),
	write('Happen.  These may happen at same time.'),nl,nl,nl.

backTrace(S1,S2) :-
	!,
%	display('  notHappenRecord start.'),ttynl,
	notHappenRecord,
	write('Not happen.  These cannot happen at same time.'),nl,nl,nl.

conflictSet(FS1,FS2,Conf) :-
	dataTransState(State1,Id1,P1,C1),
	FS1 = (State1,Id1),
	dataTransState(State2,Id2,P2,C2),
	FS2 = (State2,Id2),
	State1 \== State2,
	conflictSetSub(FS1,FS2),
	pathConflictCheck(P1,P2,[],Path,Pconf),
	ctlConflictCheck(C1,C2,[],Ctl,Cconf),
	getOpStatus(Pconf,Path,Cconf,Ctl,Status,Conf),
	Status == conflict.
	
conflictSetSub(S1,S2) :-
	recorded(already,(S2,S1),_),
	!,
	fail. 
conflictSetSub(S1,S2) :-
	!,
	recordz(already,(S1,S2),_).

dataTransState(State,Id,Path,Ctl) :-
	recorded(facility,(State,Id,Path,Ctl),_).
%	Path \== [].

printProcess([],_).
printProcess([X|L],Tab) :-
	X = (X1,X2),
	uniqState(X1,X2,XX),
	printSameState(XX,Tab),
	Ntab is Tab + 2,
	printProcess(L,Ntab).
printSameState([],_).
printSameState([X|L],Tab) :-
	!,
	tab(Tab),write(X),nl,
	printSameState(L,Tab).

uniqState(P,P,[P]) :- !.
uniqState(P,Q,[P,Q]) :- !.

happenRecord :-
	recorded(dependOn,((S1,S2),L),Ref),
	erase(Ref),
	happenRecordMain((S1,S2),L),
	fail.
happenRecord.

happenRecordMain(Set,L) :-
	recorded(sameTimeHappen,Set,_),
	!,
	recordList(sameTimeHappen,L).
happenRecordMain(_,_) :-
	!.
recordList(Key,[]).
recordList(Key,[X|L]):-
	recorded(Key,X,_),
	!,
%	display(('Same record already exists.')),ttynl,
	recordList(Key,L).
recordList(Key,[X|L]):-
	!,
	recordz(Key,X,_),
	recordList(Key,L).

notHappenRecord :-
	recorded(dependOn,(Set,L),Ref),
	erase(Ref),
	recordList(notHappen,L),
	fail.
notHappenRecord :-
	recorded(appear,L,Ref),
	erase(Ref),
	recordList(notHappen,L),
	fail.
notHappenRecord.

allConflictSet :-
	conflictSet(FS1,FS2,Conf),
%	display(('New conflict set has gotten.')),ttynl,
%	display(('New conflict set has gotten.      ')),
	fail.
allConflictSet.

allConfHandle:-
	cpu_time(T0),
	allConflictSet,
	cpu_time(T1),
	Time is T1 - T0,
	display(Time),display('sec'),ttynl.
