/*
	connect.pl

	search connection between two facilities 
*/

messageConnect(To,From,Path,Ctl,Status,Tab) :-
	tab(Tab),write('Search path for '),
	write(To),write(' <- '),write(From),nl,
	connectHandle(From,To,Path,Ctl,Status),
	messageConnectSub(Status,Path,Ctl,From,To,Tab).

messageConnectSub(fail,Path,Ctl,From,To,Tab) :-
	tab(Tab),write('Fail.'),write(' No connection.'),nl,
	!.
messageConnectSub(noOther,Path,Ctl,From,To,Tab) :-
	tab(Tab),write('No other route.'),nl,
	!.
messageConnectSub(success,Path,Ctl,From,To,Tab) :-
	tab(Tab),write('Success.'),write(' Path is '),write(Path),nl.
						% when back track , print Retry
messageConnectSub(success,Path,Ctl,From,To,Tab) :-
	tab(Tab),write('Retry for '),write(To),write(' <- '),write(From),nl,
	fail.				

connectHandle(From,To,P,C,Status) :- 
	newConnect(From,To,[],P,[],C),
	P \== error,
	C \== error,
	recordz(semaphore,(To,From),_),
	Status = success.
connectHandle(From,To,P,C,Status) :-
	recorded(semaphore,(To,From),_),
	!,
	Status = noOther,
	P = error,
	C = error.
connectHandle(From,To,P,C,Status) :-
	!,
	Status = fail,
	P = error,
	C = error.

newConnectHandle(Froms,To,Upath,Path,Uctl,Ctl) :- 
	separateBit(To,Froms,Set),
	newConnectList(Set,Upath,Path,Uctl,Ctl).

newConnectList([],Path,Path,Ctl,Ctl) :- !.
newConnectList([One|LL],Upath,Path,Uctl,Ctl) :-
	One = (To,From),
	!,
	newConnect(From,To,Upath,Tpath,Uctl,Tctl),
	newConnectList(LL,Tpath,Path,Tctl,Ctl).

newConnect(From,To,Upath,Path,Uctl,Ctl) :- 
	From = [F,[Fout,[Fl,Fh]]],
	newPath(OnePath,[F,[Fout,[Fpl,Fph]]],Wport),
	Fpl =< Fl,
	Fph >= Fl,
	pathConflictCheck([[OnePath,Wport]],Upath,[],Cpath,Conf),
	Conf == [],
	Wport = [W,[Win,[Wpl,Wph]]],
	Wl is Fl - Fpl + Wpl,
	Wh is Fh - Fph + Wph,
		  % data is on [W,[Win,[Wl,Wh]]]
 		  %   but if Wph < Wh , covers only [Wl,Wph]
	newConnectSub(Cpath,Path,Uctl,Ctl,Fph,[F,[Fout,[Fl,Fh]]],Wph,[W,[Win,[Wl,Wh]]],To).

newConnectSub(Path,Path,Ctl,Ctl,Fph,[F,[Fout,[Fl,Fh]]],Wph,To,To):-
	Fph >= Fh,
	!.

newConnectSub(Cpath,Path,Uctl,Ctl,Fph,[F,[Fout,[Fl,Fh]]],Wph,To,To):-
	          %  OnePath [F,[Fout,[Fl,Fph]]] is part for [F,[Fout,[Fl,Fh]]]
	!,
	To = [T,[Tin,[Tl,Th]]],
	Tph is Tl + Fph - Fl,
	Fnb is Fph + 1,
	Tnb is Tph + 1,
	newConnect([F,[Fout,[Fnb,Fh]]],[T,[Tin,[Tnb,Th]]],Cpath,Path,Uctl,Ctl).

newConnectSub(Cpath,Path,Uctl,Ctl,Fph,[F,[Fout,[Fl,Fh]]],Wph,[W,[Win,[Wl,Wh]]],To):-
	Fph >= Fh,
	          %  OnePath is sufficient for full [F,[Fout,[Fl,Fh]]]
	!,
	ctlLineForCnt(W,Win,Wl,Wh,OutPort,Uctl,Cctl),
	Cctl \== error,
	newConnectHandle(OutPort,To,Cpath,Path,Cctl,Ctl).

newConnectSub(Cpath,Path,Uctl,Ctl,Fph,[F,[Fout,[Fl,Fh]]],Wph,[W,[Win,[Wl,Wh]]],To):-
	          %  OnePath [F,[Fout,[Fl,Fph]]] is part for [F,[Fout,[Fl,Fh]]]
	!,
	ctlLineForCnt(W,Win,Wl,Wph,OutPort,Uctl,Cctl),
%	ctlConflictCheck(Conn,Uctl,[],Cctl,CtlConf),
%	CtlConf == [],
	Cctl \== error,
	To = [T,[Tin,[Tl,Th]]],
	Tph is Tl + Fph - Fl,
	Fnb is Fph + 1,
	Tnb is Tph + 1,
	newConnectHandle(OutPort,[T,[Tin,[Tl,Tph]]],Cpath,Dpath,Cctl,Dctl),
	newConnect([F,[Fout,[Fnb,Fh]]],[T,[Tin,[Tnb,Th]]],Dpath,Path,Dctl,Ctl).

ctlLineForCnt(W,Win,Wl,Wh,OutPort,Uctl,Cctl) :-
	newType(Fa,W),
	newFunc([conn,[Fa,[Wout,[Wol,Woh]]],[Fa,[Win,[Wcl,Wch]]]],Cconn),
	faTypeToName(Cconn,Fa,W,Conn),
	ctlConflictCheck(Conn,Uctl,[],Tctl,Conf),
	Conf == [],
	Wcl =< Wl,
	Wch >= Wl,
	Ol is Wl - Wcl + Wol,
	Oh is Wh - Wch + Woh,
			% data is on [Ol,Oh]
			%   but if Woh < Oh , covers only [Ol,Woh]
	ctlLineForCntSub([W,[Wout,[Ol,Oh]]],Win,Wch,Woh,Wh,OutPort,Tctl,Cctl).

ctlLineForCntSub([W,[Wout,[Ol,Oh]]],Win,Wch,Woh,Wh,OutPort,Ctl,Ctl) :-
	Wch >= Wh,
	!,
	OutPort = [ [W,[Wout,[Ol,Oh]]] ].
ctlLineForCntSub([W,[Wout,[Ol,Oh]]],Win,Wch,Woh,Wh,OutPort,Uctl,Cctl) :-
	!,
	Wnb is Wch + 1,
	ctlLineForCnt(W,Win,Wnb,Wh,Hport,Uctl,Cctl),
%	Hport = [W,[Wout,[Hl,Hh]]],
%	Onb is Woh + 1,
%	Onb == Hl,             % bit renzoku siteiru
                               % wakaretara dousuru???
	OutPort = [ [W,[Wout,[Ol,Woh]]] | Hport ].

faTypeToName(none,_,_,[]) :-
	!.
faTypeToName([Fa,Cnt],Fa,W,[W,Cnt]) :- 
	!.

localMessage(To,From,Status,Tab) :-
	Status = success,
	tab(Tab),write('Destination Local Variable exists in '),
	write(To),write(' <- '),write(From),nl,
	tab(Tab),write('Let Path = [].'),nl.
localMessage(To,From,noOther,Tab) :-
	!.

constantMessage(To,From,Status,Tab) :-
	Status = success,
	tab(Tab),write('Constant exists in '),
	write(To),write(' <- '),write(From),nl,
	tab(Tab),write('Let Path = [].'),nl.
constantMessage(To,From,noOther,Tab) :-
	!.

