%
%	forward.pl : time tracer    forward
%
sameTimeCheck([],_) :- 
	!,
	write(('Execution finished normally by [].')),nl.
sameTimeCheck(T,close) :- !,
	printSameTime(T,0),
	write(('Execution finished normally by close.')),nl.
sameTimeCheck(T,_) :-
	recursionCheck(T,includingRecursion),
	!,
	write(('Recursion. So back track other branch.')),nl,
	fail.			% back track
sameTimeCheck(T0,_) :-
	stateCallCheck(T0,T1),
	recordz(same_time,T1,_),
	printSameTime(T1,0),
	groundCheck([],T1,G),
	write(('Ground : ')),write(G),nl,
	recordz(ground_subint,G,_),
	oneClock(T1,T2),
	closeCheck(T2,T3,State),
	sameTimeCheck(T3,State).

groundCheckHandle(_,[],[]) :- !.
groundCheckHandle(P,[X|L],OO) :-
	!,
	groundCheck(P,X,XX),
	groundCheckHandle(P,L,LL),
	append(XX,LL,OO).

groundCheck(_,[],[]) :- !.
groundCheck(P,[X|L0],Out) :-
	!,
	X = ((PredId,Cnum,Snum),Clock,RC),	
	Y = ((PredId,Cnum,Snum),Clock),
	makeCaller(Y,P,Caller),
	groundCheckHandle(Caller,L0,L1),
	groundCheckOne(P,Y,L1,Out).

makeCaller(Y,P,Out) :-
	Y = ((PredId,Cnum,Snum),Clock),
	Snum == 1,
	Clock == 0,
	!,
	Out = [Y|P].
makeCaller(Y,P,Out) :-
	!,
	Out = [Y].
	
groundCheckOne(Caller,SS,L,Out) :-
	SS = (SubintId,0),
	recorded(facility,(State,Id,Path,Ctl),_),
	State = [SS|CC],
	append(CC,_,Caller),
	!,
	Out = [[SS|CC]|L].
groundCheckOne(P,SS,L,L) :- !.

recursionCheck(Callers,R) :-
	nowCheck(Callers,L),
	append(A,B,L),
	B \== [],
	append(BB,BB,B),
	!,
	R = includingRecursion.
recursionCheck(Callers,notRecursive) :-
	!.

nowCheckHandle([],[]) :- !.
nowCheckHandle(L,C) :-
	!,
	member(One,L),
	nowCheck(One,C).
	
nowCheck([],[]) :- !.
nowCheck([X|L],[XX|C]) :-
	X = ((PredName,Cnum,SubintNum),Clock,RC),
	!,
	XX = (PredName,Cnum),
	nowCheckHandle(L,C).

getSIState(plus,cannot) :- !. 
getSIState(notMinus,can) :- !. 
getSIState(0,close) :- !. 
getSIState(_,cannot) :- !. 

makeCloseState(cannot,cannot,cannot) :- !.
makeCloseState(can,S,S) :- !. 
makeCloseState(S,can,S) :- !. 
makeCloseState(close,close,close) :- !.

closeCheckHandle([],[],can).
closeCheckHandle([X|L],[Y|LL],State) :-
	closeCheck(X,Y,S1),
	closeCheckHandle(L,LL,S2),
	makeCloseState(S1,S2,State).

closeCheck([],[],can) :- !.
closeCheck([X|L],Out,State) :-
	closeCheckHandle(L,LL,S1),
	(
	S1 \== conflict,
		X = ((PredName,Cnum,SubintNum),Clock,RC),
		getSIState(RC,St),
		makeCloseState(S1,St,S2),
		closeCheckOne([X|LL],S2,Out,State)
	;
	S1 == conflict,
		Out = LL,
		State = conflict
	).

closeCheckOne(R,cannot,R,cannot) :- !.
closeCheckOne(R,conflict,R,conflict) :- 
	!,
	write(('This contains conflict.')),write(R),nl,
	fail.
closeCheckOne([S|_],close,Out,St) :-
	S = ((PredName,Cnum,SubintNum),Clock,RC),
	recorded(predicate,((PredName,Cnum),_,_,NS),_),
	(
	NS == SubintNum,
		Out = [],
		St = close
	;
	NS \== SubintNum,
	Nsubint is SubintNum + 1,
		recorded(sub_int,((PredName,Cnum,Nsubint),_,_,L,_),_),
		Out = [((PredName,Cnum,Nsubint),0,L)],
		(
		NS == Nsubint,
			getSIState(L,St)
		;
		NS \== Nsubint,
			St = cannot
		)
	),
	!.

closeCheckOne([S|SS],can,Out,St) :-
	S = ((PredName,Cnum,SubintNum),Clock,RC),
	recorded(predicate,((PredName,Cnum),_,_,NS),_),
	(
	NS == SubintNum,
%	Out = [S|SS],
		Out = [],
		St = can
	;
	NS \== SubintNum,
	Nsubint is SubintNum + 1,
		recorded(sub_int,((PredName,Cnum,Nsubint),_,_,L,_),_),
		Out = [((PredName,Cnum,Nsubint),0,L)],
		(
		NS == Nsubint,
			getSIState(L,St)
		;
		NS \== Nsubint,
			St = cannot
		)
	),
	!.

oneClockHandle([],[]).

oneClockHandle([X|L0],[Y|L1]) :-
	oneClock(X,Y),
	oneClockHandle(L0,L1).
	
oneClock([],[]) :- !.
oneClock([X|L0],[Y|L1]) :-
	oneClockOne(X,Y),
	oneClockHandle(L0,L1),
	!.
	
oneClockOne(SS,TT) :-
	SS = (SubintId,Clock,St),
	Nclock is Clock + 1,
	(
	integer(St),
		Nst is St - 1
	;
	St == plus,
		Nst = notMinus
	;
	St == notMinus,
		Nst = notMinus
	;
	St = Nst
	),
	TT = (SubintId,Nclock,Nst),
	!.

stateCallCheckHandle([],[]) :- !.
stateCallCheckHandle([X|L],[Y|M]) :-
	!,
	stateCallCheckHandle(L,M),
	stateCallCheck(X,Y).
	
stateCallCheck([X|L],Out) :-
	!,
	stateCallCheckHandle(L,LL),
	stateCallCheckOne(X,Xcalls),	% Xcalls do not include X
	append([X|Xcalls],LL,Out).

stateCallCheckOne(X,Xcalls) :-
	X = (SubintId,Clock,State),
	recordz(max_id,(SubintId,Clock,1),_),
	takeAllStateCall(SubintId,Clock),
	!,
	takeAllPredId(SubintId,Clock,1,Xcalls).

takeAllStateCall(SubintId,Clock) :-
	recorded(new_call,(SubintId,Clock,_),Ref),
	!.
takeAllStateCall(SubintId,Clock) :-
	recorded(state_call,(SubintId,Clock,_,PredName,Arg),_),
	recorded(max_id,(SubintId,Clock,Id),Ref),
	recordz(new_call,(SubintId,Clock,Id,PredName),_),
	erase(Ref),
	Nid is Id + 1,
	recordz(max_id,(SubintId,Clock,Nid),_),
	fail.
takeAllStateCall(SubintId,Clock) :-
	recorded(max_id,(SubintId,Clock,Id),Ref),
	recordz(new_call,(SubintId,Clock,Id,end),_),
	erase(Ref).
	
takeAllPredId(SubintId,Clock,Id,[]) :-
	recorded(new_call,(SubintId,Clock,Id,end),_),
	!.
takeAllPredId(SubintId,Clock,Id,[SSS|Out]) :-
	recorded(new_call,(SubintId,Clock,Id,PredName),_),
	!,
	recorded(predicate,((PredName,ClauseNum),A,C,NumOfSubint),_),
	recorded(sub_int,((PredName,ClauseNum,1),_,_,L,_),_),
%	determine LL from L
	LL = L,
	SS = ((PredName,ClauseNum,1),0,LL),
	stateCallCheck([SS],SSS),
	Nid is Id + 1,
	takeAllPredId(SubintId,Clock,Nid,Out).

allPrintSameTime :-
	recorded(same_time,X,_),
	printSameTime(X,0),
	fail.
allPrintSameTime.

printSameTimeHandle([],_).
printSameTimeHandle([X|L0],Tab) :-
	printSameTime(X,Tab),
	printSameTimeHandle(L0,Tab).
	
printSameTime([],_) :- !.
printSameTime([X|L0],Tab) :-
	Tab1 is Tab + 2,
	printSameTimeOne(X,Tab),
	printSameTimeHandle(L0,Tab1),
	!.
printSameTimeOne(X,Tab) :-
	tab(Tab),write(X),nl.
 
happenAtSameTime :-
	clear_record(new_call),
	clear_record(max_id),
	StartSubint = ('start/0',1,1),
	recorded(sub_int,(StartSubint,_,_,L,_),_),
	getSIState(L,St),
	sameTimeCheck([(StartSubint,0,L)],St),
	fail.

happenAtSameTime :-
	true.

groundConflictCheck :-
	nl,nl,
	write('Start Conflict Check '),
	write('between two State that happens at same time.'),nl,
	recorded(ground_subint,SS,_),
	gConfMain(SS),
	fail.
groundConflictCheck.

gConfMain(SS) :-
	uniqCombi(SS,S1,S2),
	conflictCheckState(S1,S2),
	nl,
	fail.
gConfMain(SS).

uniqCombi(SS,S1,S2) :-
	member(S1,SS),
	member(S2,SS),
	S1 \== S2,
	uniqCombiSub(S1,S2).

uniqCombiSub(S1,S2) :-
	recorded(already,(S2,S1),_),
	!,
	fail.
uniqCombiSub(S1,S2) :-
	!,
	recordz(already,(S1,S2),_).

conflictCheckState(S1,S2) :-
	write('This set happens at same time.'),nl,
	write(S1),write(' and '),write(S2),nl,
	write(' Check conflict.'),nl,
	!,
	recorded(facility,(S1,Id1,P1,C1),_),
	recorded(facility,(S2,Id2,P2,C2),_),
	pathConflictCheck(P1,P2,[],Path,Pconf),
	ctlConflictCheck(C1,C2,[],Ctl,Cconf),
	getOpStatus(Pconf,Path,Cconf,Ctl,Status,Conf),
	recordConf(S1,S2,Id1,Id2,Status).

recordConf(S1,S2,Id1,Id2,conflict) :-
	!,
	FS1 = (S1,Id1),
	FS2 = (S2,Id2),
	write('Conflict. There is conflict in this set.'),nl,
	write('   '),write(FS1),write(' and '),write(FS2),nl,
	recordz(conflict,(FS1,FS2),_),
	recordz(stateConflict,(S1,S2),_).
recordConf(S1,S2,Id1,Id2,_) :-
	!,
	FS1 = (S1,Id1),
	FS2 = (S2,Id2),
	write('No conflict. There is no conflict in this set.'),nl,
	write('   '),write(FS1),write(' and '),write(FS2),nl.

