/*
	happenSameTime.pl

    		check whether set of two subintervals happens at same time
*/

prevState([((PredName,Cnum,Snum),Clock)|Callers],Prev,SimplePrevious,Cond) :-
	integer(Clock),
	Clock \== 0,
	!,
	Pclock is Clock - 1,
	SP = ((PredName,Cnum,Snum),Pclock),
	SimplePrevious = [SP|Callers],
	sameState(SP,Callers ,Prev,Cond).

sameStateHandle([X|Callers],Out,Cond) :-
	!,
	sameState(X,Callers,Out,Cond).

sameState(Same,Callers,[Same|Callers],[]).
sameState( ((PredName,Cnum,Snum),Clock) ,Callers,Same,[]) :-
						%   go previous subinterval
	Clock == 0,
	Snum \== 1,
	!,
	Psnum is Snum - 1,
	sameStateSub( (PredName,Cnum,Psnum) ,RS),
	reverse(RS,CC),
	append(CC,Callers,Same).
sameState( ((PredName,Cnum,Snum),Clock) ,[],Same,Cond) :-
						%   may go anywhere
	Clock == 0,
	Snum == 1,
	recorded(state_call,(_,_,_,PredName,_),_),
	!,
	recorded(state_call,(CsubintId,Cclock,_,PredName,_),_),
	recorded(predicate,((PredName,Cnum),Arg,Cond1,NS),_),
	sameState((CsubintId,Cclock),[],Same,Cond2),
	append(Cond1,Cond2,Cond).
sameState( ((PredName,Cnum,Snum),Clock) ,[],Same,[]) :-
						%   no one calls PredName
	Clock == 0,
	Snum == 1,
	!,
	Same = initialState.
sameState( ((PredName,Cnum,Snum),Clock) ,Callers,Same,Cond) :-
						%   back to Callers
	Clock == 0,
	Snum == 1,
	!,
	recorded(predicate,((PredName,Cnum),Arg,Cond1,NS),_),
	sameStateHandle(Callers,Same,Cond2),
	append(Cond1,Cond2,Cond).

sameStateSub(SubintId,Same) :-
						%   fixed length
	recorded(sub_int,(SubintId,_,_,L,NDT),_),
	integer(L),
	!,
	Same = [(SubintId,L)].
sameStateSub(SubintId,Same) :-
	recorded(state_call,(SubintId,Clock,_,PredName,_),_),
	!,
	recorded(predicate,((PredName,Cnum),Arg,Cond,NumOfSubint),_),
	sameStateSub((PredName,Cnum,NumOfSubint),SS),
	Same = [(SubintId,Clock)|SS].
sameStateSub(SubintId,Same) :-
						%  idling subint
	!,
	Same = [(SubintId,1)].

happenSameTime(S1,S2,Appear,_,_,_,_) :-
	recorded(notHappen,(T1,T2),_),
	append(T1,_,S1),
	append(T2,_,S2),
	!,
	write('Recorded in notHappen.'),nl,
	write('  Self is   '),write((S1,S2)),nl,
	write('  Record is '),write((T1,T2)),nl,
	recordz(appear,Appear,_),
	fail.
happenSameTime(S1,S2,Appear,_,_,_,_) :-
	recorded(notHappen,(T1,T2),_),
	append(T1,_,S2),
	append(T2,_,S1),
	!,
	write('Recorded in notHappen.'),nl,
	write('  Self is   '),write((S1,S2)),nl,
	write('  Record is '),write((T2,T1)),nl,
	recordz(appear,Appear,_),
	fail.
happenSameTime(S1,S2,Appear,_,_,_,_) :-
						%   resursion
 	member2((S1,S2),Appear,[],LL),
	!,
	recordz(dependOn,((S1,S2),LL),_),
	write('Appeared in this parent.'),nl,
	fail.
happenSameTime(S1,S2,Appear,_,_,_,_) :-
 	member2((S2,S1),Appear,[],LL),
	!,
	write('Appeared in this parent.'),nl,
	recordz(dependOn,((S2,S1),LL),_),
	fail.
/*
happenSameTime(S1,S2,Appear,_,_,_,_) :-
	recorded(dependOn,(Set,LL),_),
	recordedAsDepend((S1,S2),LL,Appear),
	!,
	write('Recorded in dependOn.'),nl,
	fail.
*/
happenSameTime(S1,S2,Appear,_,_,_,_) :-
	S1 == initialState,
	!,
	write('Reach to intialState.'),nl,
	recordz(appear,Appear,_),
	fail.
happenSameTime(S1,S2,Appear,_,_,_,_) :-
	S2 == initialState,
	!,
	write('Reach to intialState.'),nl,
	recordz(appear,Appear,_),
	fail.
happenSameTime(S1,S2,_,Pr1,Pr2,Pr1,Pr2) :-
	recorded(sameTimeHappen,(S1,S2),_),
	!.
happenSameTime(S1,S2,_,Pr1,Pr2,Pr1,Pr2) :-
	recorded(sameTimeHappen,(S2,S1),_),
	!.
happenSameTime(S1,S2,_,Pr1,Pr2,Pr1,Pr2) :-
	append(S1,_,S2),
	!.
happenSameTime(S1,S2,_,Pr1,Pr2,Pr1,Pr2) :-
	append(S2,_,S1),
	!.
happenSameTime(S1,S2,Appear,Pr1,Pr2,Ps1,Ps2) :-
	!,
	prevState(S1,P1,SP1,Cond1),
	prevState(S2,P2,SP2,Cond2),
	notExCond(Cond1,Cond2),
	happenSameTime(P1,P2,[(S1,S2)|Appear],
				[(P1,SP1)|Pr1],[(P2,SP2)|Pr2],Ps1,Ps2),
	recordz(sameTimeHappen,(S1,S2),_).


notExCond([],Cond2).
notExCond([CC|Cond1],Cond2) :- 
	notExCondSub(CC,Cond2),
	notExCond(Cond1,Cond2),
	!.
notExCondSub(CC,[]). 
notExCondSub(CC,[SS|Cond2]) :- 
	\+ exCondOne(CC,SS).
	

exCondOne(C1,C2) :-
	notConditionOne(C1,N1),
	C2 == N1.

recordedAsDepend((S1,S2),LL,Appear):-
	member2((S1,S2),LL,[],_),
	recordz(dependOn,((S1,S2),Appear),_).
recordedAsDepend((S1,S2),LL,Appear):-
	member2((S2,S1),LL,[],_),
	recordz(dependOn,((S2,S1),Appear),_).

member2((C1,C2),[(T1,T2)|L],LL,LL) :-
	append(T1,_,C1),
	append(T2,_,C2),
	!.
member2(X,[XX|L],L0,LL) :-
	member2(X,L,[XX|L0],LL).
