/*
	pathSearch.pl

	     Search path that implements the data transfer including operation.
*/


pathSearch(To,From,Path,Ctl,Status,Tab) :-
	pathSearchSrc(To,From,Path,Ctl,Status,Tab).

pathSearchSrc(To,From,[],[],Status,Tab) :-   % Destination Local Var
	From = [&ImagReg,[local,[L,H]]],
	!,
	recorded(data_equal,(SubintId,(Dest,[&,ImagReg]),Clock),_),
	localMessage(To,(&ImagReg,((SubintId,Clock,Dest))),Status,Tab).
pathSearchSrc(To,From,[],[],Status,Tab) :-    %  Constant
	From = [Const,[const,[L,H]]],
	atomic(Const),
	!,
	constantMessage(To,Const,Status,Tab).
pathSearchSrc(To,From,Path,Ctl,Status,Tab) :-    %  port expression
	From = [F,[Fout,[Fl,Fh]]], 
	!,
	pathSearchDst(To,From,Path,Ctl,Status,Tab).
pathSearchSrc(To,*From,Path,Ctl,Status,Tab) :-    %  register
	atomic(From),
	!,
	newType(RegType,From),
	newFunc([conn,[RegType,Out],[RegType,_]],_),
	pathSearchDst(To,[From,Out],Path,Ctl,Status,Tab).
pathSearchSrc(To,From,Path,Ctl,Status,Tab) :- % op,(reg),mem,localVar,const
	!,
	srcIsOp(To,From,Path,Ctl,Status,Tab).

pathSearchDst(&To,From,[],[],Status,Tab) :-    % Destination Local Variable
	!,
	recorded(data_equal,(SubintId,(Dest,[&,To]),Clock),_),
	localMessage((&To,((SubintId,Clock,Dest))),From,Status,Tab).
pathSearchDst(To,From,Path,Ctl,Status,Tab) :-      %  To is port expression
	To = [T,[Tin,[Tl,Th]]],
	!,
	messageConnect(To,From,Path,Ctl,Status,Tab). 
pathSearchDst(*To,From,Path,Ctl,Status,Tab) :-      %  To is register
	atomic(To),
	!,
	newType(RegType,To),
	newFunc([set,[RegType,BW],[RegType,In]],_),
	messageConnect([To,In],From,Path,Ctl,Status,Tab). 

srcIsOp(To,From,Path,Ctl,Status,Tab) :-   
	tab(Tab),write('Deal with '),
	write(To),write(' <- '),write(From),nl,
	Tab1 is Tab + 2,
	seekPathSet(To,From,PathSet,Faci,Tab1),
%	Tab2 is Tab1 + 2,
	pathSetSearch(PathSet,Path,Ctl,Status,Faci,Tab1),
	srcIsOpSub2(To,From,PathSet,Path,Ctl,Status,Tab).
srcIsOp(To,From,Path,Ctl,Status,Tab) :-   
	getStatus((path,To,From),Status),
	srcIsOpSub(To,From,Status,Tab),
	Path = error,
	Ctl = error.

srcIsOpSub2(To,From,PathSet,Path,Ctl,Status,Tab) :-   
	recordz(semaphore,(path,To,From),_),
%	tab(Tab),write('Success. Path is '),write(Path),nl,
%	tab(Tab),write(' Control Line is '),write(Ctl),nl.
	tab(Tab),write('Success.'),nl.
srcIsOpSub2(To,From,PathSet,Path,Ctl,Status,Tab) :-   
	tab(Tab),write('Retry for '),
	write(To),write(' <- '),write(From),nl,
%	tab(Tab1),write('Retry for '),write(PathSet),nl,
	!,
	fail.

srcIsOpSub(To,From,Status,Tab) :-
	Status == fail,
	tab(Tab),write('Fail for '),write(To),write(' <- '),write(From),nl,
	!.
srcIsOpSub(To,From,Status,Tab) :-
	Status == noOther,
	tab(Tab),write('No other path for '),
	write(To),write(' <- '),write(From),nl,
	!.

statusCheck(Status,FF,Tab) :-
	Status == success,
	tab(Tab),write('Success.'),nl.
statusCheck(Status,FF,Tab) :-
	Status == success,
	tab(Tab),write('Retry path set for '),write(FF),nl,
	!,
	fail.
statusCheck(Status,FF,Tab) :-
	Status == fail,
	tab(Tab),write('Fail. Data path cannot implement this operation.'),nl,
%	tab(Tab),write('There is other equal operation ?'),nl,
	tab(Tab),write('Retry for other equal operation.'),nl,
	!,
	fail.
statusCheck(Status,Record,FF,Tab) :-
	Status == noOther,
	tab(Tab),write('No other path set for '),write(FF),nl,
%	tab(Tab),write('There is other equal operation ?'),nl,
	tab(Tab),write('Retry for other equal operation.'),nl,
	!,
	fail.

getOpStatus(Pconf,Path,Cconf,Ctl,Status,Conf) :-
	Path == error,
	!,
	Conf = [],
	Status = fail.
getOpStatus(Pconf,Path,Cconf,Ctl,Status,Conf) :-
	Ctl == error,
	!,
	Conf = [],
	Status = fail.
getOpStatus(Pconf,Path,Cconf,Ctl,Status,Conf) :-
	Pconf == [],
	Cconf == [],
	!,
	Conf = [],
	Status = success.
getOpStatus(Pconf,Path,Cconf,Ctl,Status,Conf) :-
	Pconf \== [],
	Cconf == [],
	!,
	Conf = Pconf,
	Status = conflict.
getOpStatus(Pconf,Path,Cconf,Ctl,Status,Conf) :-
	!,
	append(Pconf,Cconf,Conf),
	Status = conflict.

getStatus(Record,Status) :-
	recorded(semaphore,Record,_),
	!,
	Status = noOther.
getStatus(Record,Status) :-
	!,
	Status = fail.

opFaSub(Path,Ctl,Conf,Status,Record,Tab) :-
	Status == success,
	recordz(semaphore,Record,_),
	tab(Tab),write('Success. Path is '),write(Path),nl,
	tab(Tab),write('         Control Line is '),write(Ctl),nl.
opFaSub(Path,Ctl,Conf,Status,Record,Tab) :-
	Status == conflict,
	tab(Tab),write(('Conflict. Facility conflict happens in  ')),
	write(Conf),nl,
	fail.
opFaSub(Path,Ctl,Conf,Status,Record,Tab) :-
	Status == fail,
	!,
	tab(Tab),write('Fail confirmed, but search other partial success.'),nl,
	fail.
opFaSub(Path,Ctl,Conf,Status,Record,Tab) :-
	tab(Tab),write('Retry for each path.'),nl,
	fail.

