%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Tokio Interpreter (freeze version)
%%%%                               Shinji KONO @ csl.sony.co.jp
%%%% 				Fri Aug 16 00:56:39 1991 BST
%%%% Framed Version		Thu Aug 22 19:33:33 1991 BST
%%%% Projection Version		Sat Sep 14 09:57:08 1991 BST
%%%% $Header: /usr/user/kono/ITL/tokio/RCS/frame.pl,v 1.6 91/09/17 12:45:24 kono Exp Locker: kono $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


:-op(1200, fy	,[(('$define')),(('$function'))]).	/* macro definitions */
:-op(1200, xfy	,[('$clause'),('$call')]).	/* macro development */
% :-op(1100, xfy	,[( '|' )]).			/* guard */
:-op(1160, fx   ,[(tokio)]).	/* Tokio formula  interpreter */
:-op(1150, fy   , [(if),(while)]).
:-op(1150, xfy  , [(else)]). :-op(1150, xfy  , [(do)]).
:-op(1150, xfy  , (then)). 
:-op(1150, xfy  , [('&@'),('&&'),( & ),('||'),('//')]).
:-op(900,  fx   , ['<>',#,'|a|',beg,halt]).	
:-op(700,  xfy , [\=,===,' = '] ).
:-op(700,  xfy , [:=] ).		/* static assignment instanteanous */
:-op(700,  xfy , [<=] ).		/* static temporal assignment  */
:-op(700,  xfy , [<-] ).		/* temporal assignment */
:-op(700,  xfy , [<<-] ).		/* multiple temporal assignment */
:-op(700,  xfy , [<--] ).		/* force constrant */
:-op(700,  xfy , gets ).		/* repeatedly assignment */
:-op(700,  xfy , proj ).		/* projection */
:-op(600,  xfx , to ).			/* range */
:-op(150,  fy  , @).			/* next oprator */
:-op(140,  fx  , *).			/* variables */

%%%% Top level
tokio :- read(X),tokio(X).
tokio :- tokio.

tokio(X) :- tokio_clock(X,_,0,_Fin,_Out).

%%%% Clock-wise reduction

tokio_clock(true,_,Now,Fin,output) :-!,
	closing(Fin,Now).
tokio_clock(X,more,Now,Fin,output) :-!,
	increment_time(Now,Next),
	tokio(X,Xn,true,
		[Empty,Now,Fin,Out],   		% new interval data
		[0,[0]],_),			% Frame Flags [Pid,Parents]
	interval(Empty,Now,Fin),		% interval constraints
	% !,					% for no time backtrack
	tokio_clock(Xn,Empty,Next,Fin,Out).
tokio_clock(_,empty,_,_,output) :-!.   			% Discard Next Queue
	
%%%% Inteval constraints
%%%% Now is always fixed % Non-Deterministic
interval(empty,N,N).
interval(more,Now,Fin) :- nonvar(Fin),!,Fin>Now.
interval(more,_,_).
% interval(more,Now,Fin) :- freeze(Fin,Fin>Now).  % too expensive

%%%% Clock display stuff
closing(Fin,_):-var(Fin),!,nl,write('t+').
closing(Fin,Now):-nonvar(Fin),Fin>Now+1,!,increment_time(Fin,_).
closing(_,_).

increment_time(Now,Next) :- 
	Next is Now+1,
	nl,put("t"),write(Now),put(":"),ttyflush.
% Just for fun
%increment_time(Now,_) :- 
%	write('[A'),put(13),write('[K'),
%	put("t"),write(Now),put(":"),ttyflush,fail.
increment_time(Now,_) :- 
	nl,put("b"),write(Now),put(":"),ttyflush,fail.

%%%% Main tableau expansion

tokio(true,X,X,_,F,F):-!.   % including var case
tokio(!,X,X,_,F,F):-!.  % ignore ..
tokio((A,B),Xn,Xn1,Clock,F,F1) :-!,
	tokio(A,Xn, Xn0,Clock,F,F0),
	tokio(B,Xn0,Xn1,Clock,F0,F1).
tokio(empty,X,X,[Empty,Now,Fin|_],F,F):-!,Empty=empty,Now=Fin.
tokio(more,X,X,[Empty|_],F,F):-!,Empty=more.
%%%% local negation
tokio(not(A),X,X1,Clock,F,F1):-tokio(A,X,X1,Clock,F,F1),!,fail.
tokio(not(_),X,X,_,F,F):-!.
%%%% In case of meta call, top level functor have to be time constant.
tokio('$t'(A,Af,Ax),X,X1,Clock,F,F):-!,
	copy_term(A,B),unifyAll(B,'$t'(A,Af,Ax)),
	tokio(B,X,X1,Clock,F,F).
tokio(call(A),X,X1,Clock,F,F1):-!, tokio(A,X,X1,Clock,F,F1).
tokio((A;_),X,X1,Clock,F,F1):-   tokio(A,X,X1,Clock,F,F1).
tokio((_;B),X,X1,Clock,F,F1):-!, tokio(B,X,X1,Clock,F,F1).
tokio(next(A),(Ax,X),X,_,F,F):-!, unifyNowNext(A,_,_,Ax).
tokio(@A,(Ax,X),X,[Empty|_],F,F):-!,Empty=more, unifyNowNext(A,_,_,Ax).
tokio(ifMore(A),X,X1,[Empty|Clock],F,F1):-!,
	freeze(Empty,ifMore(Empty,A,X,X1,Clock,F,F1)).
    ifMore(more, A,X,X1,Clock,F,F1) :-!,
	    tokio(A,X,X1,[more|Clock],F,F1).
    ifMore(empty,_A,X,X,_,F,F).
tokio(ifEmpty(A),X,X1,[Empty|Clock],F,F1):-!,
	freeze(Empty,ifEmpty(Empty,A,X,X1,Clock,F,F1)).
    ifEmpty(empty, A,X,X1,Clock,F,F1) :-!,
	    tokio(A,X,X1,[empty|Clock],F,F1).
    ifEmpty(more, _A,X,X,_,F,F).
tokio(length(N),X,X,[Empty,Now,Fin|_],F,F):-!,
	unifyNowNext(N,Nn,_,_), 
	freeze(Nn,(Fin is Now+Nn,interval(Empty,Now,Fin))).
%%% Defered IO
tokio(write(N),X,X,[_,_,_,Out|_],F,F):-!,
	unifyNowNext(N,Nn,_,_),
	freeze(Out,write(Nn)).
tokio(nl,X,X,[_,_,_,Out|_],F,F):-!,
	freeze(Out,nl).  
%%% Chop
tokio((A & B),X,X1,Clock,F,F1):-!,
	Clock = [_,Now,Fin|T],
	SubClock = [SubEmpty,Now,SubFin|T],
	copy_term(A,Ab), % better to use get_variables.
	% freeze(SubFin,freeze(Fin,SubFin=<Fin)),
	freeze(SubFin,fincheck(Fin,SubFin)),
	tokio(unifyAll(A,Ab),Y,Y0,SubClock,F,F2),!,  % Green Cut
	tokio(Ab,Y0,Y1,SubClock,F2,F0),
	chop(SubEmpty,B,Y,Y1,X,X1,SubFin,Clock,F0,F1).
    fincheck(Fin,_):-var(Fin),!.   % freezing is too expensive here
    fincheck(Fin,SubFin):-SubFin=<Fin.
tokio('$i'(A,B,SubFin),X,X1,Clock,F,F1):-!,
	Clock = [_,Now,_|T],
	tokio(A,Y,Y1,[SubEmpty,Now,SubFin|T],F,F0),
	% SubFin=<Fin already constrainted
	chop(SubEmpty,B,Y,Y1,X,X1,SubFin,Clock,F0,F1).
    chop(empty,B,_,true,X,X1,Now,Clock,F,F1):-  % discard next queue
    	    Clock   = [_,Now|_], tokio(B,X,X1,Clock,F,F1).
    chop(more, B,Y,true,('$i'(Y,Bn,SubFin),Xn),Xn,SubFin,[more,Now|_],F,F):-
    	    % freeze(SubFin,SubFin>Now),     % too expensive
    	    (nonvar(SubFin),!,SubFin>Now;true),
	    unifyNowNext(B,_,_,Bn).
%%% Projection
%     proj(P1,P2) 
%          initiate projection (separate projection environment)
%          do P2 one clock
%          projection
%             P2 empty? do nothing
%             P2 more? , Synchronize Temporal Variable 
%                  start P1 in new interval
%                  put '$p' as new pojection inteval
%     '$p' do P1 one clock
%          prj P1 empty?
%             start new projection cycle
%          prj P1 more?
%             put '$p' 
tokio(proj(P1,P2),X,X1,Clock,F,F1):-!,
	% prepare projection clock/interval and temporal variable
	Clock = [Empty,Now,_|T], PrjClock = [Empty,Now,Pfin|T],
	copy_term(P2,P20), unifyNow(P2,P20),
	tokio(P20,P2n,true,PrjClock,F,F0),
	% interval(Empty,Now,Pfin),
	projection(P1,P2,P20,P2n,X,X1,Now,Pfin,PrjClock,Clock,F0,F1).
   projection(_,_,_,_,X,X,Pnow,Pfin,
	    [empty,Pnow,Pfin|_],[empty,Now,Fin|_],F,F):-  % !,
   	Now=Fin,Pnow=Pfin.
   projection(A,B0,B1,B,('$p'(Y,An,B0n,B1n,B,SubFin,Pnext,Pfin),Xn),Xn,
   		Pnow,Pfin,_,Clock,F,F0):-
	% send variables to the future for variable synchronization
	unifyNowNext(B0,_,_,B0n), unifyNowNext(B1,_,_,B1n),
   	% start A in subinterval (same as chop)
	Clock = [more,Now,Fin|T],SubEmpty=more,
	SubClock = [SubEmpty,Now,SubFin|T], Pnext is Pnow+1,
	copy_term(A,Ab), unifyNowNext(A,_,_,An), 
	% freeze(SubFin,freeze(Fin,SubFin=<Fin)),  % too expensive again
	freeze(SubFin,fincheck(Fin,SubFin)),
	tokio(unifyAll(A,Ab),Y,Y0,SubClock,F,F1),!, 
	tokio(Ab,Y0,true,SubClock,F1,F0).
tokio('$p'(A,P1,B0,B1,B,SubFin,Pnow,Pfin),X,X1,Clock,F,F1):-!,
	Clock = [_,Now,_|T],
	tokio(A,Y,Y1,[SubEmpty,Now,SubFin|T],F,F0),
	prj(SubEmpty,P1,B0,B1,B,Y,Y1,X,X1,SubFin,Pnow,Pfin,Clock,F0,F1).
    prj(empty,P1,B0,B1,B,_,true,X,X1,Now,Pnow,Pfin,Clock,F,F1):-
	    % start new projection cycle
    	    Clock   = [_,Now,_|T],PrjClock = [Empty,Pnow,Pfin|T], 
	    % synchronize temporal variable in projection
	    unifyNow(B0,B1), tokio(B,P2n,true,PrjClock,F,F0),
	    interval(Empty,Pnow,Pfin),
	    projection(P1,B0,B1,P2n,X,X1,Pnow,Pfin,PrjClock,Clock,F0,F1).
    prj(more, P1,B0,B1,B,Y,true,
    		('$p'(Y,Pn1,B0n,B1,B,SubFin,Pnow,Pfin),Xn),Xn,
    		SubFin,Pnow,Pfin,[more,Now|_],F,F):-
	    % continue sub interval
    	    % freeze(SubFin,SubFin>Now), not good idea to do here
    	    (nonvar(SubFin),!,SubFin>Now;true),
	    % don't make B1 next, it must be preserved until next tick
    	    unifyNowNext(B0,_,_,B0n), unifyNowNext(P1,_,_,Pn1).
%%% force weak next
tokio(prefix(P),X1,X,Clock,F,F1):-!,
        tokio('$w'(P,_),X1,X,Clock,F,F1).
tokio('$w'(true,_),X,X,_,F,F):-!.
tokio('$w'(P,Fin),('$w'(Pn,Fin),X),X,Clock,F,F1):-!,
            Clock   = [Empty,Now,_|T],PreClock = [PEmpty,Now,Fin|T],
            tokio(P,Pn,true,PreClock,F,F1),
	    (Empty=empty,Empty=PEmpty;PEmpty=more).
%%%% These are definable preficates, but brought here for efficiency reason
tokio((A &@ B),X,X1,Clock,F,F1):-!,    % This makes two times slower than & @ 
	tokio((A & @B),X,X1,Clock,F,F1).
tokio((A && B),X,X1,Clock,F,F1):-!,
	tokio((more,A & B),X,X1,Clock,F,F1).
tokio(#(A),(#(Ax),X),X1,Clock,F,F1):-!, 
	unifyNowNext(A,_,_,Ax),tokio(A,X,X1,Clock,F,F1).
tokio([](A),([](Ax),X),X1,Clock,F,F1):-!, 
	unifyNowNext(A,_,_,Ax),tokio(A,X,X1,Clock,F,F1).
tokio(keep(A),(keep(Ax),X),X1,[Empty|Clock],F,F1):-!,
	unifyNowNext(A,_,_,Ax), freeze(Empty,ifMore(Empty,A,X,X1,Clock,F,F1)).
tokio(fin(A),(fin(Ax),X),X1,[Empty|Clock],F,F1):-!,
	unifyNowNext(A,_,_,Ax), freeze(Empty,ifEmpty(Empty,A,X,X1,Clock,F,F1)).
tokio(gets(A,B),(gets(Ax,Bx),X),X1,[Empty|Clock],F,F):-!,
	unifyNowNext(B,_,_,Bx), unifyNowNext(A,_,_,Ax), 
	freeze(Empty,ifMore(Empty,Ax=B,X,X1,Clock,F,F)).
tokio({A},X,X1,Clock,F,F1):-!, 
	tokio(A,X,X1,Clock,F,F1).
%%%% Temporal assignment stuff
tokio(A<--B,X,X,Clock,F,F):-   % quick time constants
	eval(B,V,Clock),nonvar(V),A=V,!.
tokio(A<--B,(stable(Ax),X),X,Clock,F,F):-!,
	A = '$t'(An,_,Ax),Ax = '$t'(An,_,_), eval(B,An,Clock).
tokio(A=B,X,X,Clock,F,F):-!,
	eval(A,VA,FA,Clock), eval(B,VB,FB,Clock),!,
	VA=VB, flag_merge(FA,FB).
tokio(A>B,X,X,Clock,F,F):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,AV>BV)).
tokio(A<B,X,X,Clock,F,F):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,AV<BV)).
tokio(A=<B,X,X,Clock,F,F):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,AV=<BV)).
tokio(A>=B,X,X,Clock,F,F):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,AV>=BV)).
%%%% Framing Staff
tokio(unifyAll(A,B),(unifyAll(Ax,Bx),X),X,_,F,F):-!,
	unifyNow(A,B), unifyNowNext(A,_,_,Ax), unifyNowNext(B,_,_,Bx).
tokio(unify(A,B),X,X,_,F,F):-!,
	unifyNow(A,B).
tokio(frame(A),(frame(Ax),X),X,[_,_,_,Out|_],F,F):-!,
	unifyNowNext(A,Value,Flag,Ax),
	freeze(Out,frame_variable(Value,Flag,Ax)). 
   frame_variable(Value,Flag,NVar) :- untouch(Flag),!,
   	   unifyNowNext(NVar,Value,_,_).
   frame_variable(_,_,_).
tokio(process(true),X,X,_Clock,F,F):-!.    	% terminating process
tokio(process(A),(process(Ax),X),X,Clock,[I,P|T],[I2,P|T1]):-!,
   	freeze(I,I1 is I+1),
	tokio(A,Ax,true,Clock,[I1,[I1|P]|T],[I2,_|T1]).
tokio(A:=B,X,X,Clock,F,F):-!,         		% nofity all ancester
	eval(B,V,Clock),F=[_,P|_],		% this is weak version
   	unifyNowNext(A,_,Flag,Ax),
   	setflag(P,Flag,change),
   	unifyNowNext(Ax,V,_,_).
tokio(readonly(A),(readonly(Ax),X),X,_,F,F):-!,   % block modification
	unifyNowNext(A,_,Flag,Ax),        %   in current process
	F=[_,[I|_]|_], umember((I,stay),Flag).
%%%% Reflective staff for Framed Variable
tokio(flag(A,FlagO),X,X,_,F,F):-!,		% flag of variable
	F=[_,P|_],
   	unifyNowNext(A,_,Flag,_),
   	setflag(P,Flag,FlagM),
   	unifyNowNext(FlagO,FlagM,_,_).
tokio(pid(PidO),X,X,_,F,F):-!,			% current process id
	F=[_,[Pid|_]|_],
   	unifyNowNext(PidO,Pid,_,_).
tokio(ppid(PidO),X,X,_,F,F):-!,			% current process groups
	F=[_,Pid|_],
   	unifyNowNext(PidO,Pid,_,_).
%%%% Framing Flag Staff
   % setflag.  set flag up ancester
   setflag([],_,_):-!.
   setflag([P|T],Flag,Value):-!,
   	umember((P,Value),Flag), setflag(T,Flag,Value).
   % uniq member
   umember((P,V),T):-freeze(P,umember0((P,V),T)).
   umember0((P,V),[(P,V1)|_]):-!,V=V1.
   umember0(M,[_|T]):-umember0(M,T).
   %	untouch(Flag),
   untouch([]):-!.
   untouch([(_,stay)|T]):-!,untouch(T). % including variable case
   untouch([(_,change)|_]):-!,fail.
%%%% System call and User call
tokio(A,X,X,_Clock,F,F) :-
	systemp(A),!,unifyNowNext(A,An,_,_),call(An).
tokio(A,Xn,Xn1,Clock,F,F1) :-
	functor(A,H,N),functor(AA,H,N), tclause(AA,B),  
	unifyAll(A,AA),
	tokio(B,Xn,Xn1,Clock,F,F1).


%%%% Temporal Unification
%%%% Extract/Construct Timed Value

unifyNowNext('$t'(N,Flag,X),N1,Flag,X1):-!,N=N1,X=X1.
unifyNowNext(X,X1,_,X1):-atomic(X),!,X=X1.	
unifyNowNext([H|L],[Hn|Ln],_,[Hx|Lx]):-!,
	unifyNowNext(H,Hn,_,Hx), unifyNowNext(L,Ln,_,Lx).
unifyNowNext((H,L),(Hn,Ln),_,(Hx,Lx)):-!,
	unifyNowNext(H,Hn,_,Hx), unifyNowNext(L,Ln,_,Lx).
unifyNowNext(F,Fn,_,Fx) :-
	functor(F,H,A), functor(Fx,H,A),functor(Fn,H,A),
	unifyNowNext(0,A,F,Fn,Fx).
unifyNowNext(A,A,_,_,_):-!.
unifyNowNext(I,A,F,Fn,Fx):-
	J is I+1,
	arg(J,F,NX),arg(J,Fn,N),arg(J,Fx,X),
	unifyNowNext(NX,N,_,X),  % drop flag for structured data
	unifyNowNext(J,A,F,Fn,Fx).

%%%% Unify now 
%%%% with flag
unifyNow('$t'(Now,Flag,_),'$t'(Now1,Flag1,_)) :-!,
	Now=Now1,
	flag_merge(Flag,Flag1).
unifyNow('$t'(Now,Flag,_),D):-!,
	unifyNowNext(D,Now,Flag,_).
unifyNow(D,'$t'(Now,Flag,_)):-!,
	unifyNowNext(D,Now,Flag,_).
unifyNow([H|L],[H1|L1]) :- !,
	unifyNow(H,H1),unifyNow(L,L1).
unifyNow((H,L),(H1,L1)) :- !,
	unifyNow(H,H1),unifyNow(L,L1).
unifyNow(Sa,Sb):-
	functor(Sa,H,N),functor(Sb,H,N),
	unifyNow_arg(N,N,Sa,Sb).
unifyNow_arg(0,_,_,_):-!.
unifyNow_arg(I,A,F,Fn):-!,
	J is I-1, arg(I,F,NX),arg(I,Fn,N),
	unifyNow(NX,N), unifyNow_arg(J,A,F,Fn).


%%%% Unify Two Temporal Term for all time
%%%% with flag
unifyAll(V,V) :- !.			% variable case
unifyAll('$t'(Now,Flag,Nxt),'$t'(Now1,Flag1,Nxt1)) :-!,
	Now=Now1,
	flag_merge(Flag,Flag1),
	unifyAll(Nxt,Nxt1).
unifyAll('$t'(Now,Flag,Nxt),D):-!,
	unifyNowNext(D,Now,Flag,Nxt1),
	unifyAll(Nxt,Nxt1).
unifyAll(D,'$t'(Now,Flag,Nxt)):-!,
	unifyNowNext(D,Now,Flag,Nxt1),
	unifyAll(Nxt,Nxt1).
unifyAll([H|L],[H1|L1]) :- !,
	unifyAll(H,H1),unifyAll(L,L1).
unifyAll((H,L),(H1,L1)) :- !,
	unifyAll(H,H1),unifyAll(L,L1).
unifyAll(Sa,Sb):-
	functor(Sa,H,N),functor(Sb,H,N),
	unify_arg(N,N,Sa,Sb).
unify_arg(0,_,_,_):-!.
unify_arg(I,A,F,Fn):-!,
	J is I-1, arg(I,F,NX),arg(I,Fn,N),
	unifyAll(NX,N), unify_arg(J,A,F,Fn).

flag_merge(A,B):-var(A),tail(B,A),!.
flag_merge([],[]) :- !.
flag_merge([H|T],V) :- !,umember(H,V),!,
   	flag_merge(T,V).
tail(V,V1) :- var(V),!,V=V1.
tail([],[]):-!.
tail([_|T],V) :- tail(T,V).

%%%% Funtion evaluation in temporal assignment
%%%% with no flag inheritance

eval(A,V,C) :- eval(A,V,_,C).

%%%% with flag inheritance
% V is always plain variable

eval(A,V,F,_Clock):-var(A),!,unifyNowNext(A,V,F,_).
eval(A,V,_,_Clock):-number(A),!,A=V.
eval([H|T],V,_,Clock):-!,V=[HV|TV],eval(H,HV,_,Clock),eval(T,TV,_,Clock).
eval('$t'(Now,Flag,_),Now,Flag,_Clock):-!.
eval({A},V,F,_Clock):-!,unifyNowNext(A,V,F,_).  % quote
eval(now,Now,_,Clock):-!, Clock = [_,Now,_|_].
eval(fin,Fin,_,Clock):-!, Clock = [_,_,Fin|_].
eval(@A,V,F,Clock):-!,
	unifyNowNext(A,_,_,Ax), 
	eval(Ax,V,F,Clock).
eval(A+B,V,_,Clock):-!,
	eval(A,AV,_,Clock), eval(B,BV,_,Clock),
	freeze(AV,freeze(BV,(V is AV+BV))).
eval(A-B,V,_,Clock):-!,
	eval(A,AV,_,Clock), eval(B,BV,_,Clock),
	freeze(AV,freeze(BV,(V is AV-BV))).
eval(A*B,V,_,Clock):-!,
	eval(A,AV,_,Clock), eval(B,BV,_,Clock),
	freeze(AV,freeze(BV,(V is AV*BV))).
eval(A/B,V,_,Clock):-!,
	eval(A,AV,_,Clock), eval(B,BV,_,Clock),
	freeze(AV,freeze(BV,(V is AV/BV))).
eval(A mod B,V,_,Clock):-!,
	eval(A,AV,_,Clock), eval(B,BV,_,Clock),
	freeze(AV,freeze(BV,(V is AV mod BV))).
eval(F,Fn,_,Clock) :-functor(F,H,A),functor(Fn,H,A),
	evals(0,A,F,Fn,Clock).
evals(A,A,_,_,_):-!.
evals(I,A,F,Fn,C):-
	J is I+1,
	arg(J,F,NX),arg(J,Fn,N),
	eval(NX,N,_,C),
	evals(J,A,F,Fn,C).
% eval(A,V,F,_Clock):-unifyNowNext(A,V,F,_).

systemp(fail):-!.
systemp(read(_)):-!. systemp(atomic(_)):-!. systemp(functor(_,_,_)):-!.

%%%% Program Management

ld(File) :- abolish(tclause,2),
	assert(term_expansion((A:-B),tclause(A,B))),
	compile([basic|File]), retract(term_expansion(_,_)).
ls :- tclause(X,Y),numbervars((X,Y),0,_),print((X:-Y)),put("."),nl,fail. ls.
ls(P) :- tclause(X,Y),functor(X,P,_),numbervars((X,Y),0,_),
	print((X:-Y)),put("."),nl,fail. ls(_).

cputime(X) :- statistics(runtime,[X1,_]),!,X is X1/1000.
run(X) :- cputime(T0),tokio(X),cputime(T1),
	T is T1-T0,nl,write(T),write(' sec.').

?-ld([]).
% end %
