%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Tokio Interpreter (freeze version)
%%%%                               Shinji KONO
%%%% 				Fri Aug 16 00:56:39 1991 BST
%%%% Projection Version		Sat Sep 14 09:57:08 1991 BST
%%%% $Header: /usr/user/kono/ITL/tokio/RCS/tokio.pl,v 1.4 93/05/22 18:41:04 kono Exp Locker: kono $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-op(1200, fy	,[(('$define')),(('$function'))]).	/* macro definitions */
:-op(1200, xfy	,[('$clause'),('$call')]).	/* macro development */
:-op(1100, xfy	,[( '|' )]).			/* guard */
:-op(1160, fx   ,[(tokio)]).	/* Tokio formula  interpreter */
:-op(1150, fy   , [(if),(while)]).
:-op(1150, xfy  , [(else)]). :-op(1150, xfy  , [(do)]).
:-op(1150, xfy  , (then)). :-op(1150, xfy  , [('&@'),('&&'),( & ),('//')]).
:-op(900,  fx   , ['<>',#,'|a|','|t|','||',beg,halt]).	
:-op(700,  xfy , [\=,===,' = '] ).
:-op(700,  xfy , [:=] ).		/* static assignment instanteanous */
% :-op(700,  xfy , [<=] ).		/* static temporal assignment  */
:-op(700,  xfy , [<-] ).		/* temporal assignment */
:-op(700,  xfy , [<<-] ).		/* multiple temporal assignment */
:-op(700,  xfy , [<--] ).		/* force constrant */
:-op(700,  xfy , gets ).		/* repeatedly assignment */
:-op(700,  xfy , proj ).		/* projection */
:-op(600,  xfx , to ).			/* range */
:-op(150,  fy  , @).			/* next oprator */
:-op(140,  fx  , *).			/* variables */

%%%% Top level
tokio :- read(X),tokio(X).
tokio :- tokio.

tokio(X) :- tokio_clock(X,_,0,_Fin,_).

%%%% Clock wise reduction

tokio_clock(true,_,Now,Fin,true) :-!,
	closing(Fin,Now).
tokio_clock(X,more,Now,Fin,true) :-!,
	increment_time(Now,Next),
	tokio(X,Xn,true,[Empty,Now,Fin,Out]),   % new empty flag
	interval(Empty,Now,Fin),		% interval constraints
	% !,					% for no time backtrack
	tokio_clock(Xn,Empty,Next,Fin,Out).	% activate output and next
tokio_clock(_,empty,_,_,true) :-!.   		% Discard Next Queue
	
%%%% Inteval constraints
%%%% Now is always fixed % Non-Deterministic
interval(empty,N,N).
interval(more,Now,Fin) :- nonvar(Fin),!,Fin>Now.
interval(more,_,_).
% interval(more,Now,Fin) :- freeze(Fin,Fin>Now).  % not good for long interval

%%%% Clock display stuff
%closing(Fin,_):-var(Fin),!,nl,write('t+').
closing(Fin,_):-var(Fin),!,Fin=1073741824,nl,write('t+'). % big enough?
closing(Fin,Now):-nonvar(Fin),Fin>Now+1,!,increment_time(Fin,_).
closing(_,_).

increment_time(Now,Next) :- 
	Next is Now+1,
	nl,put("t"),write(Now),put(":"),ttyflush.
%%%% Just for fun :-)
increment_time(Now,_) :- 
	write('[A'),put(13),write('[K'),
	put("t"),write(Now),put(":"),ttyflush,fail.
%increment_time(Now,_) :- 
%	nl,put("b"),write(Now),put(":"),ttyflush,fail.

%%%% Main tableau expansion
tokio(true,X,X,_):-!.   % including var case
tokio(!,X,X,_):-!.  % ignore ..
tokio((A,B),Xn,Xn1,Clock) :-!,
	tokio(A,Xn, Xn0,Clock),
	tokio(B,Xn0,Xn1,Clock).
tokio(empty,X,X,[Empty,Now,Fin|_]):-!,Empty=empty,Now=Fin.
tokio(more,X,X,[Empty|_]):-!,Empty=more.
%%%% local negation
tokio(not(A),X,X1,Clock):-tokio(A,X,X1,Clock),!,fail.
tokio(not(_),X,X,_):-!.
%%%% In case of meta call, top level functor have to be time constant.
tokio('$t'(An,Ax),X,X1,Clock):-!,
	copy_term(An,B),unifyAll(B,'$t'(An,Ax)),
	tokio(B,X,X1,Clock).
tokio(call(A),X,X1,Clock):-!,
	tokio(A,X,X1,Clock).
tokio((A;_),X,X1,Clock):-
	tokio(A,X,X1,Clock).
tokio((_;B),X,X1,Clock):-!,
	tokio(B,X,X1,Clock).
tokio(next(A),(Ax,X),X,_):-!,
	unifyNowNext(A,_,Ax).
tokio(@A,(Ax,X),X,[Empty|_]):-!,Empty=more,
	unifyNowNext(A,_,Ax).
tokio(ifMore(A),X,X1,[Empty|Clock]):-!,
	freeze(Empty,ifMore(Empty,A,X,X1,Clock)).
    ifMore(more, A,X,X1,Clock) :-!,
	    tokio(A,X,X1,[more|Clock]).
    ifMore(empty,_A,X,X,_).
tokio(ifEmpty(A),X,X1,[Empty|Clock]):-!,
	freeze(Empty,ifEmpty(Empty,A,X,X1,Clock)).
    ifEmpty(empty, A,X,X1,Clock) :-!,
	    tokio(A,X,X1,[empty|Clock]).
    ifEmpty(more, _A,X,X,_).
tokio(length(N),X,X,[Empty,Now,Fin|_]):-!,
	unifyNowNext(N,Nn,_), 
	freeze(Nn,(Fin is Now+Nn,interval(Empty,Now,Fin))).
tokio(write(N),X,X,[_,_,_,Out|_]):-!,
	unifyNowNext(N,Nn,_),
	freeze(Out,write(Nn)).  % defferd until the end of this clock
tokio(nl,X,X,[_,_,_,Out|_]):-!,
	freeze(Out,nl).         % defferd until the end of this clock
tokio((A & B),X,X1,Clock):-!,
	Clock = [_,Now,Fin|T],
	SubClock = [SubEmpty,Now,SubFin|T],
	copy_term(A,Ab), % better to use get_variables.
	% This copy_term duplicates freezed predicates
	% freeze(SubFin,freeze(Fin,SubFin=<Fin)),
	tokio(unifyAll(A,Ab),Y,Y0,SubClock),!,  % Green Cut
	tokio(Ab,Y0,Y1,SubClock),
	freeze(SubFin,fincheck(Fin,SubFin)),
	chop(SubEmpty,B,Y,Y1,X,X1,SubFin,Clock).
    fincheck(Fin,_):-var(Fin),!.
    fincheck(Fin,SubFin):-SubFin=<Fin.
tokio('$i'(A,B,SubFin),X,X1,Clock):-!,
	Clock = [_,Now,_|T],
	tokio(A,Y,Y1,[SubEmpty,Now,SubFin|T]),
	% SubFin=<Fin already constrainted
	chop(SubEmpty,B,Y,Y1,X,X1,SubFin,Clock).
    chop(empty,B,_,true,X,X1,Now,Clock):-  % discard next queue
    	    Clock   = [_,Now|_], tokio(B,X,X1,Clock).
    chop(more, B,Y,true,('$i'(Y,Bn,SubFin),Xn),Xn,SubFin,[more,Now|_]):-
    	    % freeze(SubFin,SubFin>Now),
    	    (nonvar(SubFin),!,SubFin>Now;true),
	    unifyNowNext(B,_,Bn).
%%% Projection
%     proj(P1,P2) 
%          initiate projection (separate projection environment)
%          do P2 one clock
%          projection
%             P2 empty? do nothing
%             P2 more? , Synchronize Temporal Variable 
%                  start P1 in new interval
%                  put '$p' as new pojection interval
%     '$p' do P1 one clock
%          prj P1 empty?
%             start new projection cycle
%          prj P1 more?
%             put '$p' 
tokio(proj(P1,P2),X,X1,Clock):-!,
	% prepare projection clock/interval and temporal variable
	Clock = [Empty,Now,_|T], PrjClock = [Empty,Now,Pfin|T],
	copy_term(P2,P20), unifyNowNext(P2,N,_), unifyNowNext(P20,N,_),
	tokio(P20,P2n,true,PrjClock),
	projection(P1,P2,P20,P2n,X,X1,Now,Pfin,PrjClock,Clock).
   projection(_,_,_,_,X,X,Pnow,Pfin,
	    [empty,Pnow,Pfin|_],[empty,Now,Fin|_]):- % !,
   	Now=Fin,Pnow=Pfin.
   projection(A,B0,B1,B,('$p'(Y,An,B0n,B1n,B,SubFin,Pnext,Pfin),Xn),Xn,
   		Pnow,Pfin,_,Clock):-
   	% start A in subinterval (same as chop)
	Clock = [more,Now,Fin|T],SubEmpty=more,
	SubClock = [SubEmpty,Now,SubFin|T], Pnext is Pnow+1,
	copy_term(A,Ab), unifyNowNext(A,_,An), 
	% freeze(SubFin,freeze(Fin,SubFin=<Fin)),
	freeze(SubFin,fincheck(Fin,SubFin)),
	tokio(unifyAll(A,Ab),Y,Y0,SubClock),!, 
	% synchronize temporal variable in projection
	unifyNowNext(B0,_,B0n), unifyNowNext(B1,_,B1n),!,
	tokio(Ab,Y0,true,SubClock).
tokio('$p'(A,P1,B0,B1,B,SubFin,Pnow,Pfin),X,X1,Clock):-!,
	Clock = [_,Now,_|T],
	tokio(A,Y,Y1,[SubEmpty,Now,SubFin|T]),
	prj(SubEmpty,P1,B0,B1,B,Y,Y1,X,X1,SubFin,Pnow,Pfin,Clock).
    prj(empty,P1,B0,B1,B,_,true,X,X1,Now,Pnow,Pfin,Clock):-
	    % start new projection cycle
    	    Clock   = [_,Now,_|T],PrjClock = [Empty,Pnow,Pfin|T], 
    	    unifyNowNext(B0,B0n,_), unifyNowNext(B1,B0n,_),
    	    tokio(B,P2n,true,PrjClock),
	    interval(Empty,Pnow,Pfin),
	    projection(P1,B0,B1,P2n,X,X1,Pnow,Pfin,PrjClock,Clock).
    prj(more, P1,B0,B1,B,Y,true,
    		('$p'(Y,Pn1,B0n,B1,B,SubFin,Pnow,Pfin),Xn),Xn,
    		SubFin,Pnow,Pfin,[more,Now|_]):-
	    % continue sub interval
    	    % freeze(SubFin,SubFin>Now), not good idea to do here
    	    (nonvar(SubFin),!,SubFin>Now;true),
    	    unifyNowNext(B0,_,B0n), unifyNowNext(P1,_,Pn1).
%%% force weak next
tokio(prefix(P),X1,X,Clock):-!,
	tokio('$w'(P,_),X1,X,Clock).
tokio('$w'(true,_),X,X,_):-!.
tokio('$w'(P,Fin),('$w'(Pn,Fin),X),X,Clock):-!,
            Clock   = [Empty,Now,_|T],PreClock = [PEmpty,Now,Fin|T],
            tokio(P,Pn,true,PreClock),
            (Empty=empty,Empty=PEmpty;PEmpty=more).
%%%% These are definable preficates, but brought here for efficiency reason
tokio(unifyAll(A,B),(unifyAll(Ax,Bx),X),X,_):-!,
	unifyNowNext(A,V,Ax), unifyNowNext(B,V,Bx).
tokio(unify(A,B),X,X,_):-!,
	unifyNowNext(A,V,_), unifyNowNext(B,V,_).
tokio((A &@ B),X,X1,Clock):-!,    % This makes two times slower than & @ 
	tokio((A & @B),X,X1,Clock).
tokio((A && B),X,X1,Clock):-!,
	tokio(((more,A) & B),X,X1,Clock).
tokio(#(A),(#(Ax),X),X1,Clock):-!, 
	unifyNowNext(A,_,Ax),tokio(A,X,X1,Clock).
tokio([](A),([](Ax),X),X1,Clock):-!, 
	unifyNowNext(A,_,Ax),tokio(A,X,X1,Clock).
tokio(keep(A),(keep(Ax),X),X1,[Empty|Clock]):-!,
	unifyNowNext(A,_,Ax), freeze(Empty,ifMore(Empty,A,X,X1,Clock)).
tokio(fin(A),(fin(Ax),X),X1,[Empty|Clock]):-!,
	unifyNowNext(A,_,Ax), freeze(Empty,ifEmpty(Empty,A,X,X1,Clock)).
tokio(gets(A,B),(gets(Ax,Bx),X),X1,[Empty|Clock]):-!,
	unifyNowNext(B,_,Bx), unifyNowNext(A,_,Ax), 
	freeze(Empty,ifMore(Empty,Ax=B,X,X1,Clock)).
tokio({A},X,X1,Clock):-!, 
	tokio(A,X,X1,Clock).
%%%% Temporal assignment stuff
tokio(A<--B,X,X,Clock):-   % quick time constants
	eval(B,V,Clock),nonvar(V),A=V,!.
tokio(A<--B,(stable(Ax),X),X,Clock):-!,
	A = '$t'(An,Ax),Ax = '$t'(An,_), eval(B,An,Clock).
tokio(A=B,X,X,Clock):-!,
	eval(A,V,Clock), eval(B,V,Clock).
tokio(A>B,X,X,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,AV>BV)).
tokio(A<B,X,X,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,AV<BV)).
tokio(A=<B,X,X,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,AV=<BV)).
tokio(A>=B,X,X,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,AV>=BV)).
%%%% static variable
tokio(*A:=B,X,X,Clock):-!,nonvar(A),
	eval(A,AV,Clock), % calcurate arguments
	eval(B,BV,Clock),Clock = [_,_,_,Out|_],
	freeze(Out,set_static(AV,BV,Clock)).
   set_static(AV,BV,C) :-
	recorded(AV,(AV,V,C),_),!,V=BV;			% same value
	recorda(AV,(AV,BV,C),_);			% assign
	recorded(AV,(AV,_,C),Ref),erase(Ref),fail.	% backtrack
   static([]).
   static([H|L]):-
	recorded(H,_,Ref),erase(Ref),fail; static(L).
%%%% system predicate
tokio(A,X,X,_Clock) :-
	systemp(A),!,unifyNowNext(A,An,_),call(user:An).
%%%% predicate call
tokio(A,Xn,Xn1,Clock) :-
	functor(A,H,N),functor(AA,H,N), tclause(AA,B),  
	unifyAll(A,AA),
	tokio(B,Xn,Xn1,Clock).

%%%% Temporal Unification
%%%% Extract/Construct Timed Value

unifyNowNext('$t'(Now,Next),Now1,Next1):-!,Now=Now1,Next=Next1.
unifyNowNext('$i'(A, B, SubFin),
	     '$i'(An,Bn,SubFin),
	     '$i'(Ax,Bx,SubFin)):-!,
	     unifyNowNext(A,An,Ax),
	     unifyNowNext(B,Bn,Bx).
unifyNowNext('$p'(A, P1, B0, B1, B, SubFin,Pnow,Pfin),
             '$p'(An,P1n,B0n,B1n,Bn,SubFin,Pnow,Pfin),
             '$p'(Ax,P1x,B0x,B1x,Bx,SubFin,Pnow,Pfin)) :-!,
	     unifyNowNext(A,An,Ax),
	     unifyNowNext(P1,P1n,P1x),
	     unifyNowNext(B0,B0n,B0x),
	     unifyNowNext(B1,B1n,B1x),
	     unifyNowNext(B,Bn,Bx).
unifyNowNext('$p'(P,Fin), '$p'(Pn,Fin), '$p'(Px,Fin)):-!,
	     unifyNowNext(P,Pn,Px).
unifyNowNext(X,X1,X1):-atomic(X),!,X=X1.	
unifyNowNext([H|L],[Hn|Ln],[Hx|Lx]):-!,
	unifyNowNext(H,Hn,Hx), unifyNowNext(L,Ln,Lx).
unifyNowNext((H,L),(Hn,Ln),(Hx,Lx)):-!,
	unifyNowNext(H,Hn,Hx), unifyNowNext(L,Ln,Lx).
unifyNowNext(F,Fn,Fx) :-
	functor(F,H,A), functor(Fx,H,A),functor(Fn,H,A),
	unifyNowNext(0,A,F,Fn,Fx).
unifyNowNext(A,A,_,_,_):-!.
unifyNowNext(I,A,F,Fn,Fx):-
	J is I+1,
	arg(J,F,NX),arg(J,Fn,N),arg(J,Fx,X),
	unifyNowNext(NX,N,X),
	unifyNowNext(J,A,F,Fn,Fx).

%%%% Unify Two Temporal Term
unifyAll(V,V) :- !.				% variable case
unifyAll('$t'(Now,Nxt),'$t'(Now,Nxt1)) :-!,
	unifyAll(Nxt,Nxt1).
unifyAll('$t'(Now,Nxt),D):-!,
	unifyNowNext(D,Now,Nxt1),
	unifyAll(Nxt,Nxt1).
unifyAll(D,'$t'(Now,Nxt)):-!,
	unifyNowNext(D,Now,Nxt1),
	unifyAll(Nxt,Nxt1).
unifyAll([H|L],[H1|L1]) :- !,
	unifyAll(H,H1),unifyAll(L,L1).
unifyAll((H,L),(H1,L1)) :- !,
	unifyAll(H,H1),unifyAll(L,L1).
unifyAll(Sa,Sb):-
	functor(Sa,H,N),functor(Sb,H,N),
	unify_arg(N,N,Sa,Sb).
unify_arg(0,_,_,_):-!.
unify_arg(I,A,F,Fn):-!,
	J is I-1, arg(I,F,NX),arg(I,Fn,N),
	unifyAll(NX,N), unify_arg(J,A,F,Fn).

%%%% Funtion evaluation in temporal assignment
eval(A,V,_Clock):-var(A),!,unifyNowNext(A,V,_).
eval(A,V,_Clock):-number(A),!,A=V.
eval({A},V,_Clock):-!,unifyNowNext(A,V,_).  % quote
eval(now,Now,Clock):-!, Clock = [_,Now,_|_].
eval(fin,Fin,Clock):-!, Clock = [_,_,Fin|_].
eval(A,V,_Clock):-atomic(A),!,A=V.
eval('$t'(Now,_),Now,_Clock):-!.
eval(@A,V,Clock):-!,
	unifyNowNext(A,_,Ax), 
	eval(Ax,V,Clock).
eval(*A,V,Clock):-!,
	eval(A,An,Clock),
	recorded(An,(An,V,_),_),!.
eval(A+B,V,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,(V is AV+BV))).
eval(A-B,V,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,(V is AV-BV))).
eval(A*B,V,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,(V is AV*BV))).
eval(A/B,V,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,(V is AV/BV))).
eval(A mod B,V,Clock):-!,
	eval(A,AV,Clock), eval(B,BV,Clock),
	freeze(AV,freeze(BV,(V is AV mod BV))).
eval(F,Fn,Clock) :-functor(F,H,A),functor(Fn,H,A),
	evals(0,A,F,Fn,Clock).
evals(A,A,_,_,_):-!.
evals(I,A,F,Fn,C):-
	J is I+1,
	arg(J,F,NX),arg(J,Fn,N),
	eval(NX,N,C),
	evals(J,A,F,Fn,C).
% eval(A,V,_Clock):-unifyNowNext(A,V,_).

systemp(fail):-!.
systemp(read(_)):-!. systemp(atomic(_)):-!. systemp(functor(_,_,_)):-!.
systemp(arg(_,_,_)):-!.
systemp(static(_)):-!.
systemp(is(_,_)):-!.
systemp(trace):-!.
% for gmlib
systemp(=>(_,_)):-!.
systemp(<=(_,_)):-!.
systemp(nextevent(_)):-!.
systemp(waitevent(_)):-!.

%%%% Program Management

ld(File) :- abolish(tclause,2),
	assert(user:term_expansion((A:-B),tclause(A,B))),
	compile([basic|File]), retract(user:term_expansion(_,_)).
ls :- tclause(X,Y),numbervars((X,Y),0,_),print((X:-Y)),put("."),nl,fail. ls.
ls(P) :- tclause(X,Y),functor(X,P,_),numbervars((X,Y),0,_),
	print((X:-Y)),put("."),nl,fail. ls(_).

cputime(X) :- statistics(runtime,[X1,_]),!,X is X1/1000.
run(X) :- cputime(T0),tokio(X),cputime(T1),
	T is T1-T0,nl,write(T),write(' sec.').

?-ld([]).
%%%% end %
